/*
 * background.c - level background
 * Copyright (C) 2010  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "background.h"
#include "actor.h"
#include "sprite.h"

/* constants */
#define BG_MAX                  50
#define BGBEHAVIOR_MAXARGS      5

/* background behaviors */
#define BGB_DEFAULT             0
#define BGB_CIRCULAR            1
#define BGB_LINEAR              2

/* private data */
typedef struct {
    spriteinfo_t *data; /* this is not stored in the main hash */
    actor_t *actor;
    double x, y; /* useful -> circular behavior */
    int repeat_x, repeat_y;
    int behavior;
    double behavior_arg[BGBEHAVIOR_MAXARGS];
} background_t;

static int background_count; /* how many bgs? */
static background_t *background[BG_MAX]; /* bg data */

/* internal methods */
static background_t *read_background(FILE *fp); /* reads a background from fp */





/*
 * background_init()
 * Reads a background theme from a .bg file
 */
void background_init(const char *file)
{
    FILE *fp;
    char line[128], identifier[128], abs_path[1024];
    int i;

    background_count = 0;
    for(i=0; i<BG_MAX; i++)
        background[i] = NULL;

    resource_filepath(abs_path, file, sizeof(abs_path), RESFP_READ);
    logfile_message("background_init('%s')", abs_path);

    if(NULL != (fp=fopen(abs_path, "r"))) {
        while(fgets(line, 128, fp)) {

            if(sscanf(line, "%127s", identifier) < 1)
                continue;

            if(strcmp(identifier, "background") == 0) {
                logfile_message("Reading background %d...", background_count);
                background[background_count++] = read_background(fp);
                background_count = min(background_count, BG_MAX-1);
            }

        }
        fclose(fp);
        logfile_message("background_init() ok!");
    }
    else
        fatal_error("FATAL ERROR: can't read file '%s'");
}



/*
 * background_release()
 * Releases all the background data
 */
void background_release()
{
    int i;

    for(i=0; i<background_count; i++) {
        actor_destroy(background[i]->actor);
        spriteinfo_destroy(background[i]->data);
        free(background[i]);
    }

    background_count = 0;
}



/*
 * background_update()
 * Updates the bg
 */
void background_update()
{
    int i;
    double dt = game_delta();
    double t = 0.001*game_timer();

    for(i=0; i<background_count; i++) {
        /* bg movement */
        switch(background[i]->behavior) {
            case BGB_DEFAULT: {
                break;
            }

            case BGB_CIRCULAR: {
                double rx = background[i]->behavior_arg[0]; /* x-dist */
                double ry = background[i]->behavior_arg[1]; /* y-dist */
                double sx = background[i]->behavior_arg[2] * (2*PI); /* x-speed */
                double sy = background[i]->behavior_arg[3] * (2*PI); /* y-speed */
                background[i]->x = background[i]->actor->spawn_point.x + rx*cos(sx*t);
                background[i]->y = background[i]->actor->spawn_point.y + ry*sin(sy*t);
                break;
            }

            case BGB_LINEAR: {
                double sx = background[i]->behavior_arg[0]; /* x-speed */
                double sy = background[i]->behavior_arg[1]; /* y-speed */
                background[i]->x += sx*dt;
                background[i]->y += sy*dt;
                break;
            }
        }
    }
}



/*
 * background_render()
 * Renders the bg
 */
void background_render(v2d_t camera)
{
    int i;
    v2d_t topleft = v2d_new(camera.x-VIDEO_SCREEN_W/2, camera.y-VIDEO_SCREEN_H/2);
    v2d_t v = v2d_new(VIDEO_SCREEN_W/2, VIDEO_SCREEN_H/2);

    /* parallax scrolling */
    for(i=0; i<background_count; i++) {
        background[i]->actor->position.x = background[i]->x + topleft.x*background[i]->actor->speed.x;
        background[i]->actor->position.y = background[i]->y + topleft.y*background[i]->actor->speed.y;
    }

    /* render */
    for(i=0; i<background_count; i++)
        actor_render_repeat_xy(background[i]->actor, v, background[i]->repeat_x, background[i]->repeat_y);
}





/* private methods */


/* Reads a background from fp */
background_t* read_background(FILE *fp)
{
    char line[128], identifier[128], arg[128];
    double behavior_arg[BGBEHAVIOR_MAXARGS];
    int i, repeat_x=TRUE, repeat_y=TRUE, behavior=BGB_DEFAULT;
    v2d_t initial_position=v2d_new(0,0), scroll_speed=v2d_new(0,0);
    spriteinfo_t *sprite=NULL;

    for(i=0; i<BGBEHAVIOR_MAXARGS; i++)
        behavior_arg[i] = 0;

    while(fgets(line, 128, fp)) {

        if(sscanf(line, "%127s", identifier) < 1)
            continue;

        if(strcmp(identifier, "initial_position") == 0)
            sscanf(line, "%*s %lf %lf", &initial_position.x, &initial_position.y);
        else if(strcmp(identifier, "scroll_speed") == 0)
            sscanf(line, "%*s %lf %lf", &scroll_speed.x, &scroll_speed.y);
        else if(strcmp(identifier, "behavior") == 0) {
            int n;
            char *p;

            sscanf(line, "%*s %127s%n", arg, &n);
            for(p=line+n,i=0; (i<BGBEHAVIOR_MAXARGS) && (sscanf(p, "%lf%n", &behavior_arg[i], &n)!=EOF); p+=n,i++);

            if(str_icmp(arg, "DEFAULT") == 0)
                behavior = BGB_DEFAULT;
            else if(str_icmp(arg, "CIRCULAR") == 0)
                behavior = BGB_CIRCULAR;
            else if(str_icmp(arg, "LINEAR") == 0)
                behavior = BGB_LINEAR;
        }
        else if(strcmp(identifier, "repeat_x") == 0) {
            sscanf(line, "%*s %127s", arg);
            repeat_x = (str_icmp(arg, "TRUE") == 0);
        }
        else if(strcmp(identifier, "repeat_y") == 0) {
            sscanf(line, "%*s %127s", arg);
            repeat_y = (str_icmp(arg, "TRUE") == 0);
        }
        else if(strcmp(identifier, "sprite") == 0) {
            if(sprite)
                spriteinfo_destroy(sprite);

            logfile_message("Reading background sprite/animation data...");
            sprite = read_sprite(fp);
            logfile_message("done.");
        }
        else if(strcmp(identifier, "}") == 0) {
            background_t *bg = mallocx(sizeof *bg);

            bg->actor = actor_create();
            bg->data = sprite;
            bg->x = bg->actor->position.x = bg->actor->spawn_point.x = initial_position.x;
            bg->y = bg->actor->position.y = bg->actor->spawn_point.y = initial_position.y;
            bg->actor->speed.x = scroll_speed.x;
            bg->actor->speed.y = scroll_speed.y;
            bg->repeat_x = repeat_x;
            bg->repeat_y = repeat_y;
            bg->behavior = behavior;
            for(i=0; i<BGBEHAVIOR_MAXARGS; i++)
                bg->behavior_arg[i] = behavior_arg[i];

            actor_change_animation(bg->actor, sprite->animation_data[0]);
            return bg;
        }
    }

    logfile_message("Warning: read_background() returned NULL (did you forget a '}' ?)");
    return NULL;
}

