/*
 * editorgrp.c - level editor: groups
 * Copyright (C) 2010  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include "editorgrp.h"
#include "brick.h"
#include "item.h"
#include "enemy.h"

/* internal data */
#define EDITORGRP_MAX_GROUPS        500
static int group_count;
static editorgrp_entity_list_t* group[EDITORGRP_MAX_GROUPS];
static editorgrp_entity_list_t* delete_list(editorgrp_entity_list_t *list);
static editorgrp_entity_list_t* add_to_list(editorgrp_entity_list_t *list, editorgrp_entity_t entity);
static editorgrp_entity_list_t* read_group(FILE *fp);

/* imported methods */
extern int editor_is_valid_item(int item_id);


/*
 * editorgrp_init()
 * Initializes this module
 */
void editorgrp_init()
{
    int i;

    group_count = 0;
    for(i=0; i<EDITORGRP_MAX_GROUPS; i++)
        group[i] = NULL;
}


/*
 * editorgrp_release()
 * Releases this module
 */
void editorgrp_release()
{
    int i;

    for(i=0; i<group_count; i++)
        group[i] = delete_list(group[i]);
    group_count = 0;
}


/*
 * editorgrp_load_from_file()
 * Reads a list of groups from a file
 */
void editorgrp_load_from_file(const char *filename)
{
    FILE *fp;
    char abs_path[1024], line[1024];
    char word[64];
    int r;

    resource_filepath(abs_path, filename, sizeof(abs_path), RESFP_READ);
    logfile_message("editorgrp_load_from_file('%s')", filename);

    if(NULL != (fp = fopen(abs_path, "r"))) {
        while(fgets(line, 1024, fp)) {
            r = sscanf(line, "%63s", word);
            if(r > 0 && strcmp(word, "group") == 0 && group_count < EDITORGRP_MAX_GROUPS)
                group[group_count++] = read_group(fp);
        }
        fclose(fp);
    }
    else
        logfile_message("Warning: editorgrp_load_from_file('%s') - failed to open '%s' for reading", filename, abs_path);

    logfile_message("editorgrp_load_from_file() loaded %d group(s)", group_count);
}


/*
 * editorgrp_group_count()
 * How many groups there are?
 */
int editorgrp_group_count()
{
    return group_count;
}


/*
 * editorgrp_get_group()
 * Returns a group, where 0 <= id < editorgrp_group_count()
 */
editorgrp_entity_list_t* editorgrp_get_group(int id)
{
    if(group_count > 0) {
        id = clip(id, 0, group_count-1);
        return group[id];
    }
    else
        return NULL;
}






/* internal methods */


/* deletes the given linked list */
editorgrp_entity_list_t* delete_list(editorgrp_entity_list_t *list)
{
    if(list != NULL) {
        list->next = delete_list(list->next);
        free(list);
        list = NULL;
    }

    return list;
}

/* adds an entity to a list */
editorgrp_entity_list_t* add_to_list(editorgrp_entity_list_t *list, editorgrp_entity_t entity)
{
    editorgrp_entity_list_t *p = mallocx(sizeof *p);
    p->entity = entity;
    p->next = list;
    return p;
}

/* reads a group from a file */
editorgrp_entity_list_t* read_group(FILE *fp)
{
    editorgrp_entity_list_t *list = NULL;
    editorgrp_entity_t e;
    char line[128], identifier[64];
    int id, x, y;

    while(fgets(line, 128, fp)) {
        if(sscanf(line, "%63s", identifier) < 1) {
            continue;
        }
        else if(strcmp(identifier, "}") == 0) {
            return list;
        }
        else if(strcmp(identifier, "brick") == 0) {
            sscanf(line, "%*s %d %d %d", &id, &x, &y);
            e.type = EDITORGRP_ENTITY_BRICK;
            e.id = clip(id, 0, brickdata_size()-1);
            e.position = v2d_new(x,y);
            if(NULL != brickdata_get(e.id)) /* valid brick? */
                list = add_to_list(list, e);
        }
        else if(strcmp(identifier, "item") == 0) {
            sscanf(line, "%*s %d %d %d", &id, &x, &y);
            e.type = EDITORGRP_ENTITY_ITEM;
            e.id = clip(id, 0, ITEMDATA_MAX-1);
            e.position = v2d_new(x,y);
            if(editor_is_valid_item(e.id)) /* valid item? */
                list = add_to_list(list, e);
        }
        else if(strcmp(identifier, "enemy") == 0) {
            sscanf(line, "%*s %d %d %d", &id, &x, &y);
            e.type = EDITORGRP_ENTITY_ENEMY;
            e.id = clip(id, 0, ENEMYDATA_MAX-1);
            e.position = v2d_new(x,y);
            list = add_to_list(list, e);
        }
    }

    list = delete_list(list);
    logfile_message("Warning: read_group() returned NULL (did you forget a '}' ?)");
    return NULL;
}
