/*
 * main.c - main file of the game
 * Copyright (C) 2008-2010  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <allegro.h>
#include <alpng.h>
#include <string.h>
#include "util.h"
#include "scene.h"
#include "video.h"
#include "audio.h"
#include "input.h"
#include "timer.h"
#include "actor.h"
#include "brick.h"
#include "player.h"
#include "item.h"
#include "enemy.h"
#include "font.h"
#include "sprite.h"
#include "level.h"
#include "quest.h"
#include "lang.h"
#include "screenshot.h"
#include "preferences.h"


/* command line arguments */
typedef struct {
    /* video stuff */
    int video_resolution;
    int smooth_graphics;
    int fullscreen; /* fullscreen mode? */
    int color_depth; /* bits per pixel */
    int show_fps;

    /* run custom level */
    int custom_level; /* user needs to run a custom level? */
    char custom_level_path[1024]; /* filepath */

    /* run custom quest */
    int custom_quest; /* user needs to run a custom quest? */
    char custom_quest_path[1024]; /* filepath */

    /* other */
    int disable_music; /* disable music? */
    int low_memory; /* reduce memory usage */
    char language_filepath[1024];
} cmdline_t;

static cmdline_t cmdline_read(int argc, char **argv);
static void display_message(char *fmt, ...);


/* garbage collector */
static void clean_garbage();








/*
 * main()
 * Main function
 */
int main(int argc, char *argv[])
{
    scene_t *scn;
    cmdline_t cmd;

    /* Initializing Allegro */
    allegro_init();
    install_timer();
    install_keyboard();
    install_mouse();

    /* initializing the helpers */
    alpng_init();
    randomize();
    init_os_specifics();
    logfile_init();
    preferences_init();
    cmd = cmdline_read(argc, argv);

    /* initializing the managers */
    timer_init();
    video_init(cmd.video_resolution, cmd.smooth_graphics, cmd.fullscreen, cmd.color_depth);
    audio_init(cmd.disable_music);
    input_init();
    hash_init(cmd.low_memory);

    /* initializing other modules */
    storyboard_init();
    spritedata_load();
    font_init();
    screenshot_init();
    video_show_fps(cmd.show_fps);
    lang_init();
    if(strcmp(cmd.language_filepath, "") != 0)
        lang_loadfile(cmd.language_filepath);

    /* initializing game data... */
    scenestack_init();
    player_set_lives(PLAYER_INITIAL_LIVES);
    player_set_score(0);

    /* initial scene */
    if(cmd.custom_level) {
        level_setfile(cmd.custom_level_path);
        scenestack_push(storyboard_get_scene(SCENE_LEVEL));
    }
    else if(cmd.custom_quest) {
        quest_t *q = load_quest(cmd.custom_quest_path);
        quest_run(q, TRUE);
        scenestack_push(storyboard_get_scene(SCENE_QUEST));
    }
    else {
        scenestack_push(storyboard_get_scene(SCENE_INTRO));
    }

    /* main loop */
    while(!game_is_over() && !scenestack_empty()) {
        /* updating the managers */
        timer_update();
        input_update();
        audio_update();

        /* updating the current scene */
        scn = scenestack_top();
        scn->update();

        /* rendering the current scene */
        if(scn == scenestack_top()) /* scn may have been 'popped' out */
            scn->render();
        screenshot_update();
        video_render();

        /* garbage collector */
        clean_garbage();
    }

    /* releasing some modules */
    scenestack_release();
    storyboard_release();
    lang_release();
    screenshot_release();

    /* Releasing managers */
    input_release();
    video_release();
    hash_release();
    audio_release();
    timer_release();
    logfile_release();

    /* bye! */
    allegro_exit();
    return 0;
}
END_OF_MAIN()




/*
 * cmdline_read()
 * Reads the command line arguments
 */
cmdline_t cmdline_read(int argc, char **argv)
{
    int i;
    cmdline_t cmd;

    /* default values */
    cmd.video_resolution = preferences_get_videoresolution();
    cmd.smooth_graphics = preferences_get_smooth();
    cmd.fullscreen = preferences_get_fullscreen();
    cmd.show_fps = preferences_get_showfps();
    str_cpy(cmd.language_filepath, preferences_get_languagepath(), sizeof(cmd.language_filepath));
    cmd.color_depth = max(16, desktop_color_depth());
    cmd.disable_music = FALSE;
    cmd.low_memory = FALSE;
    cmd.custom_level = FALSE;
    cmd.custom_quest = FALSE;

    /* logfile */
    logfile_message("game arguments:");
    for(i=0; i<argc; i++)
        logfile_message("argv[%d]: '%s'", i, argv[i]);

    /* reading data... */
    for(i=1; i<argc; i++) {

        if(str_icmp(argv[i], "--help") == 0) {
            display_message(
                "%s usage:\n"
                "    %s [options ...]\n"
                "\n"
                "where options include:\n"
                "    --help                    displays this message\n"
                "    --version                 shows the version of this program\n"
                "    --fullscreen              fullscreen mode\n"
                "    --resolution X            sets the window size, where X = TINY (%dx%d), NORMAL (%dx%d) or MAX (adapts to your desktop)\n"
                "    --smooth                  improves the graphic quality (*)\n"
                "    --tiny                    small game window (improves the speed **). Same as --resolution TINY\n"
                "    --show-fps                 shows the FPS (frames per second) counter\n"
                "    --level \"FILEPATH\"        runs the level located at FILEPATH\n"
                "    --quest \"FILEPATH\"        runs the quest located at FILEPATH\n"
                "    --disable-music           disables all the musics\n"
                "    --color-depth X           sets the color depth to X bits/pixel, where X = 8, 16, 24 or 32 (***)\n"
                "    --language \"FILEPATH\"     sets the language file to FILEPATH (for example, %s)\n"
                "    --low-memory              advanced option - reduces the memory usage (****)\n"
                "\n"
                "(*) This option may be used to improve the graphic quality using a special algorithm.\n"
                "    You should NOT use this option on slow computers, since it may imply a severe performance hit.\n"
                "\n"
                "(**) This option should be used on slow computers.\n"
                "\n"
                "(***) Use 8 bits/pixel (256 colors) to improve even more the performance on slow computers.\n"
                "\n"
                "(****) This option should be used only if you're using a hardware with a very limited amount\n"
                "       of RAM memory (under 24MB).\n"
                "\n"
                "Please read the docs for more information.\n",
            GAME_TITLE, GAME_UNIXNAME,
            VIDEO_SCREEN_W, VIDEO_SCREEN_H, VIDEO_SCREEN_W*2, VIDEO_SCREEN_H*2,
            DEFAULT_LANGUAGE_FILEPATH);
            exit(0);
        }

        else if(str_icmp(argv[i], "--version") == 0) {
            display_message("%s %d.%d.%d\n", GAME_TITLE, GAME_VERSION, GAME_SUB_VERSION, GAME_WIP_VERSION);
            exit(0);
        }

        else if(str_icmp(argv[i], "--resolution") == 0) {
            if(++i < argc) {
                if(str_icmp(argv[i], "TINY") == 0)
                    cmd.video_resolution = VIDEORESOLUTION_1X;
                else if(str_icmp(argv[i], "NORMAL") == 0)
                    cmd.video_resolution = VIDEORESOLUTION_2X;
                else if(str_icmp(argv[i], "MAX") == 0)
                    cmd.video_resolution = VIDEORESOLUTION_MAX;
                else
                    display_message("WARNING: invalid resolution (%s).", argv[i]);
            }
        }

        else if(str_icmp(argv[i], "--smooth") == 0)
            cmd.smooth_graphics = TRUE;

        else if(str_icmp(argv[i], "--tiny") == 0)
            cmd.video_resolution = VIDEORESOLUTION_1X;

        else if(str_icmp(argv[i], "--fullscreen") == 0)
            cmd.fullscreen = TRUE;

        else if(str_icmp(argv[i], "--color-depth") == 0) {
            if(++i < argc) {
                cmd.color_depth = atoi(argv[i]);
                if(cmd.color_depth != 8 && cmd.color_depth != 16 && cmd.color_depth != 24 && cmd.color_depth != 32) {
                    display_message("WARNING: invalid color depth (%d). Changing to %d...\n", cmd.color_depth, 16);
                    cmd.color_depth = 16;
                }
            }
        }

        else if(str_icmp(argv[i], "--show-fps") == 0)
            cmd.show_fps = TRUE;

        else if(str_icmp(argv[i], "--level") == 0) {
            if(++i < argc) {
                cmd.custom_level = TRUE;
                resource_filepath(cmd.custom_level_path, argv[i], sizeof(cmd.custom_level_path), RESFP_READ);
                if(!filepath_exists(cmd.custom_level_path))
                    fatal_error("FATAL ERROR: file '%s' does not exist!\n", cmd.custom_level_path);
            }
        }

        else if(str_icmp(argv[i], "--quest") == 0) {
            if(++i < argc) {
                cmd.custom_quest = TRUE;
                resource_filepath(cmd.custom_quest_path, argv[i], sizeof(cmd.custom_quest_path), RESFP_READ);
                if(!filepath_exists(cmd.custom_quest_path))
                    fatal_error("FATAL ERROR: file '%s' does not exist!\n", cmd.custom_quest_path);
            }
        }

        else if(str_icmp(argv[i], "--language") == 0) {
            if(++i < argc) {
                resource_filepath(cmd.language_filepath, argv[i], sizeof(cmd.language_filepath), RESFP_READ);
                if(!filepath_exists(cmd.language_filepath))
                    fatal_error("FATAL ERROR: file '%s' does not exist!\n", cmd.language_filepath);
            }
        }

        else if(str_icmp(argv[i], "--disable-music") == 0)
            cmd.disable_music = TRUE;

        else if(str_icmp(argv[i], "--low-memory") == 0)
            cmd.low_memory = TRUE;

        else { /* unknown option */
            display_message("%s: bad command line option \"%s\".\nRun %s --help to get more information.\n", GAME_UNIXNAME, argv[i], GAME_UNIXNAME);
            exit(0);
        }

    }

    /* done! */
    return cmd;
}


/*
 * display_message()
 * Displays a message (printf format)
 */
void display_message(char *fmt, ...)
{
    char buf[5120];
    va_list args;

    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);

#ifdef __WIN32__
    allegro_message(buf);
#else
    puts(buf);
#endif
}


/*
 * clean_garbage()
 * Runs the garbage collector.
 * (used when low memory optimizations are enabled)
 */
void clean_garbage()
{
    static uint32 last = 0;
    uint32 t = game_timer();

    if(t >= last + 2000) { /* every 2 seconds */
        last = t;
        hash_release_unused_resources();
    }
}

